local EventManager =  FlexBar2:NewModule("EventManager", "AceEvent-2.0");
EventManager.Version = 1;

EventManager.EventTable = {};
local EventTable = EventManager.EventTable;

function EventManager:OnInitialize()
    -- Schedule an updater event so that stuff without event can update
    EventManager:ScheduleRepeatingEvent("FB2_UPDATE" , 0.3);
end

EventManager.ButtonMixin = FlexBar2.OO.Mixin
{
	"RegisterEvent",
	"UnregisterEvent",
    "TriggerEvent",
	"RegisterScript",
	"UnregisterScript",
	"UnregisterAll",
};

function EventManager.ButtonMixin:RegisterEvent(Event, MethodName)
	FlexBar2:Debug("Registering " , Event, " with " , MethodName);
    -- Sanity check
	if((type(MethodName) == "string" and type(self[MethodName]) == "function") or type(MethodName) == "function") then
		if(not EventManager:IsEventRegistered(Event)) then
			-- Create the event handling function that AceEvent will trigger, will loop through all buttons for the event and trigger all registered methods for it
			EventManager[Event] = function(...)
                -- Ignore debug output for the FB2_UPDATE event, we dont want to spam the damn chatbox
                if(Event ~= "FB2_UPDATE") then FlexBar2:Debug("Triggering ", Event); end
				-- Loop through all buttons
				for Button , EventTable in pairs(EventTable[Event]) do
					-- Loop through all methods for the event found
					for _, MethodName in pairs(EventTable) do
						if(type(MethodName) == "string") then
							Button[MethodName](Button, Event, ...);
						elseif(type(MethodName) == "function") then
							MethodName(Button, Event, ...);
						end
					end
				end
			end
			-- Actualy register the event
			EventManager:RegisterEvent(Event);
		end
		-- Now we're sure the event is registered, lets store the methodname and the button which registered it
		if(type(EventTable[Event]) ~= "table") then
			EventTable[Event] = {};
		end
		if(type(EventTable[Event][self]) ~= "table") then
			EventTable[Event][self] = {};
		end
		table.insert(EventTable[Event][self], MethodName);
	else
		-- Failure, lets spit out some debug shit
		FlexBar2:Debug("Failure to register event [" , Event, "] with [", MethodName, "] as method");
		return false;
	end
end

function EventManager.ButtonMixin:UnregisterEvent(Event, MethodName)
	-- If we had a method name passed, remove said method
	if(MethodName and Event) then
		return FlexBar2:RemoveByValue(EventTable[Event][self], MethodName);
	elseif(Event) then -- if we dont have a methodname passed but do have an event passed, remove it all
		for n,v in pairs(EventTable[Event][self]) do
			table.remove(EventTable[Event][self], n);
		end
	end
end

function EventManager.ButtonMixin:TriggerEvent(Event, ...)
    EventManager:TriggerEvent(Event, ...);
end

-- Scripts are hooked using SetScript, as to not interfer with whatever the state button's scripts are
function EventManager.ButtonMixin:RegisterScript(Trigger, MethodName, ShareWithGroup) 
	-- Make sure ShareWithGroup is not nil if omitted.
    if(ShareWithGroup == nil) then ShareWithGroup = false; end
	-- Sanity checks, if MethodName is a string, make sure the method exists. a function is also valid
    if((type(MethodName) == "string" and type(self[MethodName]) == "function") or type(MethodName) == "function") then
		if(type(self.ScriptTable) ~= "table") then
			self.ScriptTable = {};
		end
		-- SetScript a special trigger function if it hasnt already
		if(type(self.Frame:GetScript(Trigger)) ~= "function") then
			self.Frame:SetScript(Trigger, function(...)
				-- Loop through all registered methods
				for MethodName , ShareWithGroup in pairs(self.ScriptTable[Trigger]) do						
					-- Dont trigger debug code of the script is OnUpdate (which is unlikely anyway, as there is an FB2_UPDATE event around, but you never know)
					if(Trigger ~= "OnUpdate") then FlexBar2:Debug("Triggering ", MethodName, " For Handler: ", Trigger); end
					-- Wether the script should be shared with its group (this is only used in a few cases but it can be handy)
                    if(ShareWithGroup) then
						if(type(MethodName) == "string") then
							-- Loop through all group members and trigger it for them too
							for _, GroupMember in pairs(self.GroupMemberList) do
								GroupMember[MethodName](GroupMember, Trigger, ...);
							end
						elseif(type(MethodName) == "function") then
							-- Same as above, but a function has been directly passed instead of a method name
							for _, GroupMember in pairs(self.GroupMemberList) do
								MethodName(GroupMember, Trigger, ...);
							end
						end
                      else
						if(type(MethodName) == "string") then
							-- Trigger script for this button
							self[MethodName](self, Trigger, ...);
						elseif(type(MethodName) == "function") then
							-- Same as above, but a function has been directly passed instead of a method name
							MethodName(self, Trigger, ...);
						end
                    end
				end
			end);
		end
		-- Now the script handler is registered for sure, lets add the one we got passed to our list to trigger
		if(type(self.ScriptTable[Trigger]) ~= "table") then
			self.ScriptTable[Trigger] = {};
		end
		self.ScriptTable[Trigger][MethodName] = ShareWithGroup;
        FlexBar2:Debug("Registering ", Trigger, " ," , MethodName);
        return true;
	else
		-- Failed, spit out debug crap and return false
        FlexBar2:Debug("Failed at registering ", Trigger, " ," , MethodName);
		return false;
	end
end

function EventManager.ButtonMixin:UnregisterScript(Trigger, MethodName)
	-- Nothing to possibly unregister if scripttable is nil, so return;
	if(type(self.ScriptTable) ~= "table") then return; end
	-- MethodName was passed, remove said method
	if(MethodName and Trigger and self.ScriptTable[Trigger]) then
		self.ScriptTable[Trigger][MethodName] = nil;
	elseif(Trigger and self.ScriptTable[Trigger]) then -- Only trigger was passed, remove it all
		for n,v in pairs(self.ScriptTable[Trigger]) do
			table.remove(self.ScriptTable[Trigger], n);
		end
	end
end
